/*
###################################################
rncontrol.h

Diese Header-Datei stellt grundlegende Funktionen fr das RN-Control 1.4 in C zur Verfgung.

Autor: Georg Richter
#######################################################
*/


#include <util/delay.h>



/*### waitms - Programm pausieren lassen ###*/

/*Die Funktion lsst circa so viele Millisekunden verstreichen, wie angegeben werden.
Angepasst auf das RN-Control 1.4 mit 16 MHz-Quarz!
Vorsicht, Wert ist nur experimentell angenhert, nicht exakt berechnet!*/

void waitms(uint16_t ms)
{
	for(; ms>0; ms--)
	{
		uint16_t __c = 4000;
		__asm__ volatile (
			"1: sbiw %0,1" "\n\t"
			"brne 1b"
			: "=w" (__c)
			: "0" (__c)
		);
	}
}




/*### Ports setzen ###*/

//Ports auf HIGH setzen
static inline void setportaon(const uint8_t n)
{PORTA |= (1<<n);}	//set PORTA.n high

static inline void setportbon(const uint8_t n)
{PORTB |= (1<<n);}	//set PORTB.n high

static inline void setportcon(const uint8_t n)
{PORTC |= (1<<n);}	//set PORTC.n high

static inline void setportdon(const uint8_t n)
{PORTD |= (1<<n);}	//set PORTD.n high



//Ports auf LOW setzen
static inline void setportaoff(const uint8_t n)
{PORTA &= ~(1<<n);}	//set PORTA.n low

static inline void setportboff(const uint8_t n)
{PORTB &= ~(1<<n);}	//set PORTB.n low

static inline void setportcoff(const uint8_t n)
{PORTC &= ~(1<<n);}	//set PORTC.n low

static inline void setportdoff(const uint8_t n)
{PORTD &= ~(1<<n);}	//set PORTD.n low





/*### Senden per USART - RS232-Kommunikation ###*/

/*Zum senden von Zeichen im Hauptprogramm entweder

char irgendwas[] = "meintext";
sendUSART(irgendwas);

oder direkt

sendUSART("meinText");

verwenden.*/ 

void init_USART(void)
{
	UCSRB |= (1<<TXEN);	//UART TX (Transmit - senden) einschalten
	UCSRC |= (1<<URSEL)|(3<<UCSZ0);	//Modus Asynchron 8N1 (8 Datenbits, No Parity, 1 Stopbit)
	UBRRH = 0;				//Highbyte ist 0
	UBRRL = 103;	//Lowbyte ist 103 (dezimal) -> (Frequenz_in_Hz / (Baudrate * 16)) - 1 <- Quarfrequenz = 16*1000*1000 Hz!!!!
}



void sendchar(unsigned char c)
{
	while(!(UCSRA & (1<<UDRE))) //Warten, bis Senden mglich ist
	{
	}
	
	UDR = c; //schreibt das Zeichen aus 'c' auf die Schnittstelle
}



void sendUSART(char *s) //*s funktiniert wie eine Art Array - auch bei einem String werden die Zeichen (char) einzeln ausgelesen - und hier dann auf die Sendeschnittstelle bertragen
{
	while(*s)
	{
	sendchar(*s);
	s++;
	}
}





/*### ADC-Ansteuerung ###*/

uint16_t adcwert(uint8_t kanal)
{
	uint16_t wert = 0; //Variable fr Ergebnis deklarieren

	ADCSRA = (1<<ADEN)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0);	//ADEN aktiviert berhaupt erst den internen ADC-Wandler, ADPS2 bis ADPS0 stellen den verwendeten Prescaler ein, denn die Wandlerfrequenz muss immer zwischen 50 und 200 kHz liegen! Der Prescaler muss bei 16MHz also zwischen 80 und 320 eingestellt werden, als einzige Mglichkeit bleibt hier 128 (=alle auf 1).   

	ADMUX = kanal;
	//ADMUX = (1<<REFS1)|(1<<REFS0); //Einstellen der Referenzspannung auf "extern", also REFS1 und REFS0 auf "0" - daher auskommentierte Zeile
	
	ADCSRA |= (1<<ADSC);	//nach Aktivierung des ADC wird ein "Dummy-Readout" empfohlen, man liest also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen"      
    while(ADCSRA & (1<<ADSC)) {} //auf Abschluss der Konvertierung warten
	wert = ADCW;	//ADCW muss einmal gelesen werden, sonst wird Ergebnis der nchsten Wandlung nicht bernommen.
 
	/* Eigentliche Messung - Mittelwert aus 4 aufeinanderfolgenden Wandlungen */
	wert = 0; 
	for(uint8_t i=0; i<4; i++)
	{
		ADCSRA |= (1<<ADSC); 	//eine Wandlung "single conversion" starten
		while(ADCSRA & (1<<ADSC)) {} 	//auf Abschluss der Konvertierung warten
		wert = wert + ADCW;	 //Wandlungsergebnisse aufaddieren
    }
	
	ADCSRA &= ~(1<<ADEN);	//ADC deaktivieren
 
	wert = wert/4;		//Durchschnittswert bilden
 
	return wert;
}





/*### Buttonabfrage ###*/

uint8_t button(void)
{
	uint8_t taste = 0; 	//Variable fr Nummer des Tasters
	uint16_t analog7 = adcwert(7);	//Wert des Ports
	
	setportaon(7);		//Ohne das hier "flackern" die Werte aus irgend einem Grund -> es werden mitunter Tasten erkannt, die gar nicht gedrckt wurden oder das Programm bleibt fr einige Sekunden "hngen"
	waitms(1);
	setportaoff(7);
	
	//Abfrage des gedrckten Tasters - um Strungen zu vermeiden wurden
        //die Bereiche sehr eng gefasst, sollten bei Bedarf an jedes Board extra angepasst werden.
	if((analog7>=337) && (analog7<=343)) {taste = 1;}
	else if((analog7>=268) && (analog7<=274)) {taste = 2;}
	else if((analog7>=200) && (analog7<=206)) {taste = 3;}
	else if((analog7>=132) && (analog7<=138)) {taste = 4;}
	else if((analog7>=64) && (analog7<=70)) {taste = 5;}
	else {}
	
	return taste;
}




/*### Sound durch den Speaker ausgeben ###*/

/*i wird immer um das doppelte der Tonhhe hochgezhlt. Grund: Um so tiefer
der Ton, desto grer ist "hoehe" -> desto lnger sind die Pausenzeiten pro
"Tonschwingung". Dadurch wrden tiefe Tne immer lnger werden, was durch
diese Funktion abgefangen wird. Alle Tne mit der Lnge x werden auch
ca x ms lang gespielt. Vernachlssigt wird hierbei jedoch die bentigte Zeit
zum Umschalten der Pins und Hochzhlen der Schleife, um so hher der Ton, 
desto lnger wird er also real gespielt, weil fter gezhlt und umgeschaltet
werden muss. Bei hoehe=1 bentigt ein Ton etwa das 1,733-fache der Zeit, die
als Dauer angegeben wird; bei hoehe=30 wird sie hingegen beinahe eingehalten.
Der Multiplikator von 15 in der Lnge gleicht die seltsame Dauer, die die
Delay-Schleife hat, aus (die kommt aus irgend einem Grund nicht mal annhernd
an Millisekunden heran).*/


void sound(uint8_t hoehe, uint16_t laenge)
{
	for(uint16_t i=0; i<laenge*15; i=i+(2*hoehe))
	{
		setportdon(7);
		_delay_loop_2(hoehe<<7); // 0,032ms*hoehe warten bei einem 16 MHz-Quarz
		setportdoff(7);			// (hoehe<<7) ist eine effizientere Schreibweise von hoehe*128
		_delay_loop_2(hoehe<<7);
	}
}





/*### PWM-Routinen zur Motoransteuerung ###*/

void init_timer1(void)	//Initialisierung des Timers fr Erzeugung des PWM-Signals
{
   /* normale 8-bit PWM aktivieren (nicht invertiert),
   Das Bit WGM10 wird im Datenblatt auch als PWM10 bezeichnet */
   TCCR1A = (1<<COM1A1)|(1<<COM1B1)|(1<<WGM10);

   /* Einstellen der PWM-Frequenz auf 14 kHz ( Prescaler = 1 ) */
   TCCR1B = (1<<CS10);

   /* Interrupts fr Timer1 deaktivieren
   Achtung : Auch die Interrupts fr die anderen Timer stehen in diesem Register */
   TIMSK &= ~0x3c;
}



void setPWMlinks(uint8_t speed) //Geschwindigkeit linker Motor
{OCR1BL = speed;}

void setPWMrechts(uint8_t speed) //Geschwindigkeit rechter Motor
{OCR1AL = speed;}



void Mlinkszur(void)	//Uhrzeigersinn
{PORTC |= (1<<PC6); PORTC &= ~(1<<PC7);}

void Mlinksvor(void)	//mathematischer Drehsinn
{PORTC &= ~(1<<PC6); PORTC |= (1<<PC7);}

void Mlinksstop(void)	//aus
{ PORTC &= ~(1<<PC6); PORTC &= ~(1<<PC7);}



void Mrechtsvor(void)	//Uhrzeigersinn
{PORTB |= (1<<PB0); PORTB &= ~(1<<PB1);}

void Mrechtszur(void)	//mathematischer Drehsinn
{PORTB &= ~(1<<PB0); PORTB |= (1<<PB1);}

void Mrechtsstop(void)	//aus
{PORTB &= ~(1<<PB0); PORTB &= ~(1<<PB1);}